''******************************
''*  Altair CPU Simulator v1.1 *
''*  (C) 2007 Michael Rychlik  *
''******************************

' v1.0 - 01 Oct 2007 - original version
' v1.1 - 09 Dec 2007 - Moved dispatch table into HUB RAM to save COG space and increase speed
' 


CON                                
  
  'Register defs
  reg_b = 0
  reg_c = 1
  reg_d = 2
  reg_e = 3
  reg_h = 4
  reg_l = 5
  reg_m = 6
  reg_a = 7

  'Operation defs for accumulator instructions
  op_add = 0
  op_adc = 1
  op_sub = 2
  op_sbb = 3
  op_ana = 4
  op_xra = 5
  op_ora = 6
  op_cmp = 7
              
  'Register pair defs for register pair instructions
  rp_bc  = 0
  rp_de  = 1
  rp_hl  = 2
  rp_psw = 3
  rp_sp  = rp_psw
  rp_mask= %00110000


  '8080 instruction op codes

  'Carry bit instructions
  cmc   = %00111111
  stc   = %00110111
  
  'Single register instructions
  inr_mask = %11_000_111
  inr      = %00_000_100
  inr_b    = %00_000_100
  inr_c    = %00_001_100
  inr_d    = %00_010_100
  inr_e    = %00_011_100
  inr_h    = %00_100_100
  inr_l    = %00_101_100
  inr_m    = %00_110_100
  inr_a    = %00_111_100

  dcr_mask = %11_000_111
  dcr      = %00_000_101
  dcr_b    = %00_000_101
  dcr_c    = %00_001_101
  dcr_d    = %00_010_101
  dcr_e    = %00_011_101
  dcr_h    = %00_100_101
  dcr_l    = %00_101_101
  dcr_m    = %00_110_101
  dcr_a    = %00_111_101

  'data transfer instructions
  mov_mask     = %11000000
  mov_         = %01000000
  mov_src_mask = %00000111
  mov_dst_mask = %00111000

  mov_b_b      = %01_000_000
  mov_b_c      = %01_000_001
  mov_b_d      = %01_000_010
  mov_b_e      = %01_000_011
  mov_b_h      = %01_000_100
  mov_b_l      = %01_000_101
  mov_b_m      = %01_000_110
  mov_b_a      = %01_000_111

  mov_c_b      = %01_001_000
  mov_c_c      = %01_001_001
  mov_c_d      = %01_001_010
  mov_c_e      = %01_001_011
  mov_c_h      = %01_001_100
  mov_c_l      = %01_001_101
  mov_c_m      = %01_001_110
  mov_c_a      = %01_001_111

  mov_d_b      = %01_010_000
  mov_d_c      = %01_010_001
  mov_d_d      = %01_010_010    
  mov_d_e      = %01_010_011
  mov_d_h      = %01_010_100
  mov_d_l      = %01_010_101
  mov_d_m      = %01_010_110
  mov_d_a      = %01_010_111

  mov_e_b      = %01_011_000
  mov_e_c      = %01_011_001
  mov_e_d      = %01_011_010
  mov_e_e      = %01_011_011
  mov_e_h      = %01_011_100
  mov_e_l      = %01_011_101
  mov_e_m      = %01_011_110
  mov_e_a      = %01_011_111

  mov_h_b      = %01_100_000
  mov_h_c      = %01_100_001
  mov_h_d      = %01_100_010
  mov_h_e      = %01_100_011
  mov_h_h      = %01_100_100
  mov_h_l      = %01_100_101
  mov_h_m      = %01_100_110
  mov_h_a      = %01_100_111

  mov_l_b      = %01_101_000
  mov_l_c      = %01_101_001
  mov_l_d      = %01_101_010
  mov_l_e      = %01_101_011
  mov_l_h      = %01_101_100
  mov_l_l      = %01_101_101
  mov_l_m      = %01_101_110
  mov_l_a      = %01_101_111

  mov_m_b      = %01_110_000
  mov_m_c      = %01_110_001
  mov_m_d      = %01_110_010
  mov_m_e      = %01_110_011
  mov_m_h      = %01_110_100
  mov_m_l      = %01_110_101
  mov_m_m      = %01_110_110
  mov_m_a      = %01_110_111

  mov_a_b      = %01_111_000
  mov_a_c      = %01_111_001
  mov_a_d      = %01_111_010
  mov_a_e      = %01_111_011
  mov_a_h      = %01_111_100
  mov_a_l      = %01_111_101
  mov_a_m      = %01_111_110
  mov_a_a      = %01_111_111



  stax_mask    = %11101111
  stax         = %00000010
  stax_b       = %00000010
  stax_d       = %00010010

  ldax_mask    = %11101111
  ldax         = %00001010
  ldax_b       = %00001010
  ldax_d       = %00011010

  'Register or memory to accumulator instructions
  rmtoacc_mask     = %11_000_000
  rmtoacc          = %10_000_000
  rmtoacc_op_mask  = %00_111_000
  rmtoacc_reg_mask = %00_000_111
  add_b            = %10_000_000
  add_c            = %10_000_001
  add_d            = %10_000_010
  add_e            = %10_000_011
  add_h            = %10_000_100
  add_l            = %10_000_101
  add_m            = %10_000_110
  add_a            = %10_000_111
         
  adc_b            = %10_001_000
  adc_c            = %10_001_001
  adc_d            = %10_001_010
  adc_e            = %10_001_011
  adc_h            = %10_001_100
  adc_l            = %10_001_101
  adc_m            = %10_001_110
  adc_a            = %10_001_111

  sub_b            = %10_010_000
  sub_c            = %10_010_001
  sub_d            = %10_010_010
  sub_e            = %10_010_011
  sub_h            = %10_010_100
  sub_l            = %10_010_101
  sub_m            = %10_010_110
  sub_a            = %10_010_111

  sbb_b            = %10_011_000
  sbb_c            = %10_011_001
  sbb_d            = %10_011_010
  sbb_e            = %10_011_011
  sbb_h            = %10_011_100
  sbb_l            = %10_011_101
  sbb_m            = %10_011_110
  sbb_a            = %10_011_111

  ana_b            = %10_100_000
  ana_c            = %10_100_001
  ana_d            = %10_100_010
  ana_e            = %10_100_011
  ana_h            = %10_100_100
  ana_l            = %10_100_101
  ana_m            = %10_100_110
  ana_a            = %10_100_111

  xra_b            = %10_101_000
  xra_c            = %10_101_001
  xra_d            = %10_101_010
  xra_e            = %10_101_011
  xra_h            = %10_101_100
  xra_l            = %10_101_101
  xra_m            = %10_101_110
  xra_a            = %10_101_111
                                
  ora_b            = %10_110_000
  ora_c            = %10_110_001
  ora_d            = %10_110_010
  ora_e            = %10_110_011
  ora_h            = %10_110_100
  ora_l            = %10_110_101
  ora_m            = %10_110_110
  ora_a            = %10_110_111

  cmp_b            = %10_111_000
  cmp_c            = %10_111_001
  cmp_d            = %10_111_010
  cmp_e            = %10_111_011
  cmp_h            = %10_111_100
  cmp_l            = %10_111_101
  cmp_m            = %10_111_110
  cmp_a            = %10_111_111

  'Rotate accumulator instructions
  rot_mask = %111_00_111
  rot      = %000_00_111
  rlc      = %000_00_111
  rrc      = %000_01_111
  ral      = %000_10_111
  rar      = %000_11_111
  op_rlc   = 0
  op_rrc   = 1
  op_ral   = 2
  op_rar   = 3

  'PUSH instuctions
  push_mask = %11_00_1111
  push      = %11_00_0101
  push_b    = %11_00_0101
  push_d    = %11_01_0101
  push_h    = %11_10_0101
  push_psw  = %11_11_0101

  'POP instructions
  pop_mask  = %11_00_1111
  pop       = %11_00_0001
  pop_b     = %11_00_0001
  pop_d     = %11_01_0001
  pop_h     = %11_10_0001
  pop_psw   = %11_11_0001

  'DAD insructions
  dad_mask = %11_00_1111
  dad      = %00_00_1001
  dad_b    = %00_00_1001
  dad_d    = %00_01_1001
  dad_h    = %00_10_1001
  dad_sp   = %00_11_1001

  'DCX instructions
  dcx_mask = %11_00_1111
  dcx      = %00_00_1011
  dcx_b    = %00_00_1011
  dcx_d    = %00_01_1011
  dcx_h    = %00_10_1011
  dcx_sp   = %00_11_1011

  'INX instructions
  inx_mask = %11_00_1111
  inx      = %00_00_0011
  inx_b    = %00_00_0011
  inx_d    = %00_01_0011
  inx_h    = %00_10_0011
  inx_sp   = %00_11_0011

  lxi_mask = %11_00_1111
  lxi      = %00_00_0001
  lxi_b    = %00_00_0001
  lxi_d    = %00_01_0001
  lxi_h    = %00_10_0001
  lxi_sp   = %00_11_0001

  
  mvi_mask = %11_000_111
  mvi      = %00_000_110
  mvi_b    = %00_000_110
  mvi_c    = %00_001_110
  mvi_d    = %00_010_110
  mvi_e    = %00_011_110
  mvi_h    = %00_100_110
  mvi_l    = %00_101_110
  mvi_m    = %00_110_110
  mvi_a    = %00_111_110

  accimm_mask = %11_000_111
  accimm      = %11_000_110
  adi         = %11_000_110
  aci         = %11_001_110
  sui         = %11_010_110
  sbi         = %11_011_110
  ani         = %11_100_110
  xri         = %11_101_110
  ori         = %11_110_110
  cpi         = %11_111_110

  diradd_mask = %111_00_111
  diradd      = %001_00_010
  dirop_sta   = %10
  dirop_lda   = %11
  dirop_shdl  = %00
  dirop_lhdl  = %01  
  sta         = %001_10_010
  lda         = %001_11_010
  shld        = %001_00_010
  lhld        = %001_01_010  

  jmp_mask    = %11_000_111
  jmp_cond    = %11_000_010
  jmp_        = %11_000_011
  jnz         = %11_000_010
  jz          = %11_001_010
  jnc         = %11_010_010
  jc          = %11_011_010
  jpo         = %11_100_010
  jpe         = %11_101_010
  jp          = %11_110_010
  jm          = %11_111_010
  
  call_mask   = %11_000_111
  call_cond   = %11_000_100
  call_       = %11_001_101
  cnz         = %11_000_100
  cz          = %11_001_100
  cnc         = %11_010_100
  cc          = %11_011_100
  cpo         = %11_100_100
  cpe         = %11_101_100
  cp          = %11_110_100
  cm          = %11_111_100

  ret_mask    = %11_000_111
  ret_cond    = %11_000_000
  ret_        = %11_001_001
  rnz         = %11_000_000
  rz          = %11_001_000
  rnc         = %11_010_000
  rc          = %11_011_000
  rpo         = %11_100_000
  rpe         = %11_101_000
  rp          = %11_110_000
  rm          = %11_111_000

  'Condition fields for conditional jumps, calls and returns
  cond_nz     = %000
  cond_z      = %001
  cond_nc     = %010
  cond_c      = %011
  cond_po     = %100
  cond_pe     = %101
  cond_p      = %110
  cond_m      = %111

  rst_mask    = %11_000_111
  rst         = %11_000_111

  io_mask     = %11110111
  ei_di_mask  = %11110111

  
  noop  = %00000000
  halt  = %01110110

  out   = %11010011
  in    = %11011011
  di    = %11110011
  ei    = %11111011

  xchg  = %11101011
  xthl  = %11100011
  sphl  = %11111001
  
  'Jump instructions
  pchl  = %11101001

  cma   = %00101111
  daa   = %00100111

  'Program status word bits
  sign_bit   = %10000000
  zero_bit   = %01000000
  aux_bit    = %00010000
  parity_bit = %00000100
  carry_bit  = %00000001


VAR

  long  cog


PUB start(cpupar) : okay

'' Start Altair CPU simulator - starts a cog
'' returns false if no cog available
''
''   cpupar = pointer to CPU parameters

  ram_base := long[cpupar]
  ios_op   := long[cpupar + 4]
  ios_port := ios_op + 1
  ios_data := ios_op + 2

  reg_ptr  := @registers
  dispatch_tab_00_addr := @dispatch_tab_00
  
  stop
  okay := cog := cognew(@run, cpupar) + 1


PUB stop

'' Stop Altair CPU simulator - frees a cog

  if cog
    cogstop(cog~ - 1)


DAT

                        org     0
             
run
do_nop                  call    #read_ins      'Get the next instruction into 'ins'


                        mov     op, ins           'Extract 3 bits from instruction bits 3 to 5 as op
                        shr     op, #3
                        and     op, #%00000111

                        mov     reg_pair, ins     'Extract register pair code from instruction into reg_pair
                        and     reg_pair, #rp_mask
                        shr     reg_pair, #3      'N.B. shift by 3 not 4 as reg_pair indexes the first reg in register file       

                        mov     reg, ins           'Extract register number from instruction into reg
                        and     reg, #%00_000_111

                        mov     mem_ptr, h        'Default mem_ptr to content of hl regs as used by many instructions
                        shl     mem_ptr, #8
                        add     mem_ptr, l


                        mov     scratch,ins       'Is ths a MOV instruction ?
                        shr     scratch,#6
                        cmp     scratch, #%01 wz
                  if_z  jmp     #do_mov
                        cmp     scratch, #%10 wz  'Or an reg/mem to accumulator instruction ?
                  if_z  jmp     #do_rmtoacc

                        mov     wr_t, ins         'None of the above so let's find it in the dispatch table.
                        and     wr_t, #%01111111
                        shl     wr_t, #1             ' mul by 2 to get a WORD offset into the table
                        add     wr_t, dispatch_tab_00_addr
                        rdword  target, wr_t
                                
                        mov     scratch, target   'Extract some switches from the entry
                        shr     scratch, #9

                        test    scratch, #%0000001 wz  'Does this instructon need an immediate byte ?
              if_z      jmp     #no_immediate

                        call    #read_ins            'Yes: Read immediate byte into 'imm_data'
                        mov     imm_data, ins

                        test    scratch, #%0000010 wz 'Does this instruction need an immediate word ?
              if_z      jmp     #no_immediate

                        call    #read_ins            'Yes:
                        mov     mem_ptr, ins         'Immediate words are always memory reference so
                        shl     mem_ptr, #8          'build a memory pointer
                        or      mem_ptr, imm_data
no_immediate
                        and     target, #%111111111             
                        jmp     target               'Jump to the instruction handler


               
imm_data                long 0        
target                  long 0

read_ins                mov     wr_t, ram_base
                        add     wr_t, pc
                        rdbyte  ins, wr_t
                        add     pc, #1
                        and     pc, word_mask
read_ins_ret            ret
wr_t              long    0

read_mem                mov     wr_t, ram_base
                        add     wr_t, mem_ptr
                        rdbyte  alu, wr_t
read_mem_ret            ret

write_mem               mov     wr_t, ram_base
                        add     wr_t, mem_ptr
                        wrbyte  alu, wr_t
write_mem_ret           ret


do_stc                  or      psw, #carry_bit
                        jmp     #run

do_cmc                  xor     psw, #carry_bit
                        jmp     #run
           
do_daa
                        jmp     #run
       
             

do_halt                 jmp     #do_halt

do_di
do_ei
do_ei_di                mov     int_enable_flag, op
                        jmp     #run


do_in
do_out
do_io                   wrbyte  imm_data, ios_port
                        wrbyte  a, ios_data
                        wrbyte  op, ios_op
             
io_not_rdy              rdbyte  scratch, ios_op wz
              if_nz     jmp     #io_not_rdy
                        rdbyte  a, ios_data
                        jmp     #run

ios_port                long    0
ios_op                  long    0
ios_data                long    0

do_cma                  xor     a, #%11111111
                        jmp     #run


do_stax_b               mov     mem_ptr, b
                        shl     mem_ptr, #8
                        add     mem_ptr, c
                        jmp     #do_stax
           

do_stax_d               mov     mem_ptr, d
                        shl     mem_ptr, #8
                        add     mem_ptr, e
do_stax                 mov     alu, a
                        call    #write_mem
                        jmp     #run


do_ldax_b               mov     mem_ptr, b
                        shl     mem_ptr, #8
                        add     mem_ptr, c
                        jmp     #do_ldax


do_ldax_d               mov     mem_ptr, d
                        shl     mem_ptr, #8
                        add     mem_ptr, e


do_ldax                 call    #read_mem
                        mov     a, alu
                        jmp     #run


do_inr                  mov     scratch, #1          'Set up a one for increment
                        jmp     #bump


do_dcr                  mov     scratch, #%1111_1111 'Set up a minus one for decrement
bump                    mov     reg, op              'Extract register number from instruction op field
                        call    #get_reg_m          'Get specified reg from reg file or memory(hl) into alu
                        mov     op, #8               'op = 8 is alu add leaving carry alone
                        call    #alu_op             'Bump and set flags
                        call    #put_reg_m          'Put alu to specified reg in reg file or memory(hl)
                        jmp     #run


do_ral                  rol     a, #1
                        test    psw, #carry_bit wz
              if_nz     or      a, #1
                        test    a, #%1_0000_0000 wz
                        jmp     #do_ral_tail
             
do_rar                  ror     a, #1 wc
                        test    psw, #carry_bit wz
              if_nz     or      a, #%10000000
                        jmp     #do_rar_tail

do_rlc                  rol     a, #1
                        test    a, #%1_0000_0000 wz
              if_nz     or      a, #1
do_ral_tail
              if_nz     or      psw, #carry_bit    
              if_z      andn    psw, #carry_bit    
                        jmp     #run

do_rrc                  ror     a, #1 wc 
              if_c      or      a, #%10000000
do_rar_tail
              if_c      or      psw, #carry_bit    
              if_nc     andn    psw, #carry_bit
                        jmp     #run


do_mov                  cmp     ins, #halt wz       'Moving mem to mem is n ot allowed and defined as HALT
              if_z      jmp     #do_halt
                        call    #get_reg_m          'Get specified reg from reg file into alu
                        jmp     #mov_end

do_mvi                  mov     alu, imm_data       'Get the imediate data
mov_end                 mov     reg, op             'op contains dest register of mvi instructions
                        call    #put_reg_m
                        jmp     #run

do_rmtoacc              call    #get_reg_m          'Get specified reg from reg file into alu
                        mov     scratch, alu         'alu into scratch
                        jmp     #acc_end             


do_accimm
do_adi
do_aci
do_sui
do_sbi
do_ani
do_xri
do_ori
do_cpi                  mov     scratch, imm_data    'Immediate data is alu operand in scratch
acc_end                 mov     alu, a
                        call    #alu_op
                        cmp     op, #op_cmp wz
              if_nz     mov     a,alu
                        jmp     #run


do_lhld                 call    #read_mem
                        mov     l, alu
                        add     mem_ptr, #1
                        call    #read_mem
                        mov     h, alu
                        jmp     #run

do_shld                 mov     alu, l
                        call    #write_mem
                        add     mem_ptr, #1
                        mov     alu, h
                        call    #write_mem
                        jmp     #run

do_sta                  mov     alu, a
                        call    #write_mem           'Write a byte from alu into mem_ptr
                        jmp     #run

do_lda                  call    #read_mem            'Read a byte from mem_ptr into alu
                        mov     a, alu
                        jmp     #run

             
do_push                 call    #get_reg_pair        'Get specified register pair into 'scratch'
                        call    #push_word           'Push the register pair
                        jmp     #run

do_pop                  call    #pop_word            'Pop word from stack into 'scratch'
                        call    #put_reg_pair        'Put 'scratch' to specified register pair_pair
                        jmp     #run

do_dad                  call    #get_reg_pair_sp     'Get specified register pair into 'scratch'
                        mov     add_hl_t, h
                        shl     add_hl_t, #8
                        add     add_hl_t, l
                        add     add_hl_t, scratch
                        test    add_hl_t, bit_16 wz
              if_nz     or      psw, #carry_bit
                        mov     l, add_hl_t
                        and     l, #%11111111
                        shr     add_hl_t, #8
                        mov     h, add_hl_t
                        and     h, #%11111111
                        jmp     #run
                        
add_hl_t                long     0
                        

do_inx                  call    #get_reg_pair_sp        'Get specified register pair into 'scratch'
                        add     scratch, #1
                        jmp     #lxi_end

             
do_dcx                  call    #get_reg_pair_sp       'Get specified register pair into 'scratch'
                        sub     scratch, #1
                        jmp     #lxi_end


do_lxi                  mov     scratch, mem_ptr
lxi_end                 call    #put_reg_pair_sp       'Put scratch to specified register pair
                        jmp     #run

do_xchg                 mov     op, h
                        mov     scratch, l
                        mov     h, d
                        mov     l, e
                        mov     d, op
                        mov     e, scratch
                        jmp     #run

do_xthl                 mov     mem_ptr, sp
                        call    #read_mem
                        mov     scratch, l
                        mov     l, alu
                        mov     alu, scratch
                        call    #write_mem
                        add     mem_ptr, #1
                        call    #read_mem
                        mov     scratch, h
                        mov     h, alu
                        mov     alu, scratch
                        call    #write_mem
                        jmp     #run

do_sphl                 mov     sp, h
                        shl     sp, #8
                        add     sp, l
                        jmp     #run

do_pchl                 mov     pc, h
                        shl     pc, #8
                        add     pc, l
                        jmp     #run

'Conditional return instructions
do_rnz
do_rz
do_rnc
do_rc
do_rpo
do_rpe
do_rp
do_rm
do_ret_cond             movs    fytch, #do_ret
                        jmp     #do_conditional

'Conditional call instructions
do_cnz
do_cz
do_cnc
do_cc
do_cpo
do_cpe
do_cp
do_cm
do_call_cond            movs    fytch, #do_call
                        jmp     #do_conditional

'Conditional jmp instructions
do_jnz
do_jz
do_jnc
do_jc
do_jpo
do_jpe
do_jp
do_jm
do_jmp_cond             movs    fytch, #do_jmp
             
do_conditional
                        mov     scratch, psw         'Get curent state of the flags 
                        test    op, #1 wz            'wc here keeps low bit of condition indicating inverse test
              if_z      xor     scratch, #%11111111 'Invert flags for not zero, no carry, parity even and plus conditions            

                        andn    op, #1              'Remove low bit
                        shl     op, #1

                        movs    jmp_test, #test_cond
                        add     jmp_test, op
                        nop 
jmp_test                jmp     #test_cond
       
test_cond               test    scratch, #zero_bit wz
              if_nz     jmp     fytch
                        jmp     #run
                        nop         'This pads out steps between tests to 4 longs so the jumps work !!!

                        test    scratch, #carry_bit wz
              if_nz     jmp     fytch
                        jmp     #run
                        nop         'This pads out steps between tests to 4 longs so the jumps work !!!

                        test    scratch, #parity_bit wz
              if_nz     jmp     fytch
                        jmp     #run
                        nop         'This pads out steps between tests to 4 longs so the jumps work !!!

                        test    scratch, #sign_bit wz
              if_z      jmp     #run

fytch         if_nz     jmp     #do_jmp             
                        jmp     #run                

do_jmp                  mov     pc, mem_ptr         'Set program counter to the target address in immediate data 
                        jmp     #run


do_ret                  call    #pop_word           'Pop the return address from the stack into 'scratch'
                        mov     pc, scratch         'Set program counter to popped address
                        jmp     #run
                  

do_call                 mov     scratch, pc         'Push return address to stack
                        call    #push_word
                        jmp    #do_jmp


do_rst                  mov     pc, #0              'RST, Restart at PC = 0 + op field shifted up by 3
                        shl     op, #3
                        add     pc, op
                        jmp     #run



'Pop a word from the stack into 'scratch'
pop_word                mov     pp_ptr, ram_base
                        add     pp_ptr, sp
                        rdbyte  pop_tmp, pp_ptr
                        add     sp, #1
                        and     sp, word_mask
                        mov     pp_ptr, ram_base
                        add     pp_ptr, sp
                        rdbyte  scratch, pp_ptr
                        shl     scratch, #8
                        or      scratch, pop_tmp
                        add     sp, #1
                        and     sp, word_mask
pop_word_ret            ret
pop_tmp                 long    0

push_byte               sub     sp, #1               'Decrement the stack pointer
                        and     sp, word_mask
                        mov     pp_ptr, ram_base
                        add     pp_ptr, sp
                        wrbyte  scratch, pp_ptr
push_byte_ret           ret

'Push a word from 'scratch' onto the stack
push_word               ror     scratch, #8         'Get high byte into low end
                        call    #push_byte
                        rol     scratch, #8         'Get low byte back into position
                        call    #push_byte
push_word_ret           ret
pp_ptr                  long    0
             
'Perform the alu operation specified by 'op'
alu_op                  and     alu, #255                'Ensure both operands are in range
                        and     scratch, #255

                        movs    fotch, #alu_op_tbl             
                        add     fotch, op       'Make a jump to the required alu op handler
alu                     nop                     'Essential nop to cater for propeller pipeline
                                     'It also holds the 'alu' variable !!!
fotch                   jmp     #alu_op_tbl            
alu_op_tbl              jmp     #alu_add
                        jmp     #alu_adc
                        jmp     #alu_sub
                        jmp     #alu_sbb
                        jmp     #alu_and
                        jmp     #alu_xor
                        jmp     #alu_or
                        jmp     #alu_sub        'Compares (CMP & CPI) are subtracts but the result is discarded
                        'jmp     #alu_add_ncj
alu_add_ncj
                        call    #alu_add_nc
                        jmp     #alu_op_tail

                        
'Add 'scratch' and carry to 'alu' and set carry flag  accordingly
alu_adc                 test    psw, #carry_bit wz
              if_nz     add     scratch, #1

      
'Add 'scratch' to 'alu' and set carry flag and aux carry accordingly
alu_add
                        call    #alu_add_nc
              if_z      test    alu, #256 wz
              if_nz     or      psw, #carry_bit
              if_z      andn    psw, #carry_bit
                        jmp     #alu_op_tail

little_alu              long    0

                           
alu_add_nc              mov     little_alu, alu
                        add     alu, scratch
                        and     little_alu, #%1111
                        add     little_alu, scratch
                        test    little_alu, #%10000 wz
              if_nz     or      psw, #aux_bit
              if_z      and     psw, #aux_bit      
alu_add_nc_ret          ret


'Subtract 'scratch' and carry from 'alu' and set carry flag accordingly
alu_sbb                 test    psw, #carry_bit wz
              if_nz     add     scratch, #1

'Subtract 'scratch' from 'alu' and set carry flag  accordingly
alu_sub                 neg     scratch, scratch
                        jmp     #alu_add
             
'Bitwise AND of 'scratch' and 'alu' and clear carry flag
alu_and                 and     alu, scratch
                        jmp     #clear_carry

'Bitwise ExOR of 'scratch' and 'alu' and clear carry flag 
alu_xor                 xor     alu, scratch
                        jmp     #clear_carry

'Bitwise OR of 'scratch' and 'alu' and clear carry flag 
alu_or                  or      alu, scratch
clear_carry             andn    psw, #carry_bit          ' Clear the carry flag

alu_op_tail             andn    psw, #%11000100         ' Clear zero, sign and parity bits
                        and     alu, #255 wz, wc
              if_z      or      psw, #zero_bit
              if_nc     or      psw, #parity_bit
                        test    alu, #128 wz
              if_nz     or      psw, #sign_bit
alu_op_ret              ret

'Put the 'alu' to the register 'reg' in the register file or memory(mem_ptr)
put_reg_m               cmp     reg, #reg_m wz
              if_nz     jmp     #put_reg

                        call    #write_mem
                        jmp     #put_reg_m_ret     'TBD...So why not just 'ret' here (move _ret lable of course)

'Put the 'alu' to the register 'reg' in the register file
put_reg                 movd    fitch, #registers 'Use reg to index into registers
                        shl     reg, #9
                        add     fitch, reg
                        and     alu, #255          'Ensure register will be in range
fitch                   mov     registers, alu     'Get specified registers contents 
put_reg_m_ret
put_reg_ret             ret

put_reg_pair_sp         cmp     reg_pair, #%110 wz
              
              if_z      mov     sp, scratch
              if_z      jmp     #put_reg_pair_sp_ret

'Put 'scratch' into register pair specified by 'reg_pair'
put_reg_pair            cmp     reg_pair, #%110 wz    'If reg_pair is 'psw' adjust it to index A and PWS in reg file
              if_z      add     reg_pair, #1 
                        mov     reg_tmp, #registers  'Point to the register file base
                        add     reg_tmp, reg_pair    'Offset to required register pair
                        movd    put_r_high, reg_tmp 'Modify our put low byte inst. to write that reg.
                        add     reg_tmp, #1          'Point to next reg
                        movd    put_r_low, reg_tmp 'Modify our put high byte inst. to write that reg.
                        mov     reg_tmp, scratch
                        and     reg_tmp, #255
put_r_low               mov     registers, reg_tmp   'Put the low byte to it's reg
                        shr     scratch, #8          'Get the high byte
                        and     scratch, #255
put_r_high              mov     registers, scratch   'and put it to it's reg
put_reg_pair_sp_ret
put_reg_pair_ret        ret
                        

reg_tmp                 long    0

'Get the register 'reg' from the register file or memory(mem_ptr) into 'alu' 
get_reg_m               cmp     reg, #reg_m wz
              if_nz     jmp     #get_reg
                        call    #read_mem
                        jmp     #get_reg_m_ret       'TBD...So why not just 'ret' here (move _ret lable of course)


'Get the register 'reg' from the register file into 'alu' 
get_reg                 movs    fetch, #registers 'Use reg to index into registers
                        add     fetch, reg
                        nop                        'Essential nop to cater for propeller pipeline
fetch                   mov     alu, registers     'Get specified registers contents 
get_reg_m_ret
get_reg_ret             ret


get_reg_pair_sp         cmp     reg_pair, #%110 wz
              if_z      mov     scratch, sp
              if_z      jmp     #get_reg_pair_sp_ret

'Get the register pair specified by 'reg_pair' into 'scratch'
get_reg_pair
                        cmp     reg_pair, #%110 wz    'If reg_pair is 'psw' adjust it to index A and PWS in reg file
              if_z      add     reg_pair, #1 
                        mov     reg_tmp, #registers  'Point to the register file base
                        add     reg_tmp, reg_pair    'Offset to required register pair
                        movs    :get_r_high, reg_tmp 'Modify our get low byte inst. to read that reg.
                        add     reg_tmp, #1          'Point to next reg
                        movs    :get_r_low, reg_tmp 'Modify our get high byte inst. to write that reg.
:get_r_high             mov     scratch, registers    'Get the high reg and 
                        shl     scratch, #8          'move to the high byte
:get_r_low              add     scratch, registers    'Get the reg into low byte
get_reg_pair_sp_ret
get_reg_pair_ret        ret


{{
' Display a byte on the leds for one second
display      mov             dira, pins    'Set Pins to output
             shl             disp, #16     'Get data into right bits
             mov             outa, disp    'Display it
             mov             time, cnt     'Calculate delay time
             add             time, delay   'Set 1 second delay here
             waitcnt         time, 0       'Wait
display_ret  ret
disp long 0
pins      long %00000000_11111111_00000000_00000000
time    long 0
delay   long 80_000_000
}}


reg_pair                long    0


bit_16                  long    %00000000_00000001_00000000_00000000
bit_31                  long    %10000000_00000000_00000000_00000000
word_mask               long    %00000000_00000000_11111111_11111111


'8080 register file, DO NOT change the order of these or add anything in between
registers
b                       long    0
c                       long    0
d                       long    0
e                       long    0
h                       long    0
l                       long    0
m                       long    0 'Not used just holds a space so that get_reg works
a                       long    0
psw                     long    %00000000


'Other 8080 regs
sp                      long    0
pc                      long    0

'Interrupt enable flip flop
int_enable_flag         long    0


dump_regs               mov     scratch, #12              
dmp                     wrlong  registers,reg_ptr 'modify dst to get next register 
                        add     reg_ptr,#4
                        add     dmp,dst_1
                        sub     scratch,#1
                        tjnz    scratch,#dmp                             
dump_regs_ret           ret

reg_ptr  long 0
dst_1    long %1000000000
   
'Misc stuff
ins                     long    0
ram_base                long    0
scratch                 long    0
mem_ptr                 long    0
op                      long    0
reg                     long    0

dispatch_tab_00_addr    long    0

cog_end_lab

'81 longs free !!!  + There are 5 nops to use up yet !

                        fit

                        
'Instruction dispatch table contains:
'                   Address             9 bits,
'                   Switches            7 bits.
' Address is the routine that will handle this opcode
' Switches are:
'          %0000001 indicates read next code byte as immediate data for instructions that need it
'          %0000010 indicates read another code byte and make 16 bit immediate data
'          N.B. For 16 bit immediate use both switches.
'Notice that this is used from HUB RAM as it will not fit in the cog. There is a small speed penalty
'in doing this (rdword) but not much compared to total instruction execution time.
dispatch_tab_00
             '00xxxxxx instructions
{00000000}     word (%0000000 << 9) | do_nop  
{00000001}     word (%0000011 << 9) | do_lxi  
{00000010}     word (%0000000 << 9) | do_stax_b  
{00000011}     word (%0000000 << 9) | do_inx
{00000100}     word (%0000000 << 9) | do_inr
{00000101}     word (%0000000 << 9) | do_dcr  
{00000110}     word (%0000001 << 9) | do_mvi  
{00000111}     word (%0000000 << 9) | do_rlc  
{00001000}     word (%0000000 << 9) | run  
{00001001}     word (%0000000 << 9) | do_dad  
{00001010}     word (%0000000 << 9) | do_ldax_b  
{00001011}     word (%0000000 << 9) | do_dcx  
{00001100}     word (%0000000 << 9) | do_inr  
{00001101}     word (%0000000 << 9) | do_dcr  
{00001110}     word (%0000001 << 9) | do_mvi  
{00001111}     word (%0000000 << 9) | do_rrc  
{00010000}     word (%0000000 << 9) | run  
{00010001}     word (%0000011 << 9) | do_lxi  
{00010010}     word (%0000000 << 9) | do_stax_d  
{00010011}     word (%0000000 << 9) | do_inx  
{00010100}     word (%0000000 << 9) | do_inr  
{00010101}     word (%0000000 << 9) | do_dcr  
{00010110}     word (%0000001 << 9) | do_mvi  
{00010111}     word (%0000000 << 9) | do_ral  
{00011000}     word (%0000000 << 9) | run  
{00011001}     word (%0000000 << 9) | do_dad  
{00011010}     word (%0000000 << 9) | do_ldax_d  
{00011011}     word (%0000000 << 9) | do_dcx  
{00011100}     word (%0000000 << 9) | do_inr  
{00011101}     word (%0000000 << 9) | do_dcr  
{00011110}     word (%0000001 << 9) | do_mvi  
{00011111}     word (%0000000 << 9) | do_rar  
{00100000}     word (%0000000 << 9) | run  
{00100001}     word (%0000011 << 9) | do_lxi  
{00100010}     word (%0000011 << 9) | do_shld  
{00100011}     word (%0000000 << 9) | do_inx  
{00100100}     word (%0000000 << 9) | do_inr  
{00100101}     word (%0000000 << 9) | do_dcr  
{00100110}     word (%0000001 << 9) | do_mvi  
{00100111}     word (%0000000 << 9) | do_daa  
{00101000}     word (%0000000 << 9) | run  
{00101001}     word (%0000000 << 9) | do_dad  
{00101010}     word (%0000011 << 9) | do_lhld  
{00101011}     word (%0000000 << 9) | do_dcx  
{00101100}     word (%0000000 << 9) | do_inr  
{00101101}     word (%0000000 << 9) | do_dcr  
{00101110}     word (%0000001 << 9) | do_mvi  
{00101111}     word (%0000000 << 9) | do_cma  
{00110000}     word (%0000000 << 9) | run  
{00110001}     word (%0000011 << 9) | do_lxi  
{00110010}     word (%0000011 << 9) | do_sta  
{00110011}     word (%0000000 << 9) | do_inx  
{00110100}     word (%0000000 << 9) | do_inr  
{00110101}     word (%0000000 << 9) | do_dcr  
{00110110}     word (%0000001 << 9) | do_mvi  
{00110111}     word (%0000000 << 9) | do_stc  
{00111000}     word (%0000000 << 9) | run  
{00111001}     word (%0000000 << 9) | do_dad  
{00111010}     word (%0000011 << 9) | do_lda  
{00111011}     word (%0000000 << 9) | do_dcx  
{00111100}     word (%0000000 << 9) | do_inr  
{00111101}     word (%0000000 << 9) | do_dcr  
{00111110}     word (%0000001 << 9) | do_mvi  
{00111111}     word (%0000000 << 9) | do_cmc  

'''
{11000000}     word (%0000000 << 9) | do_rnz  
{11000001}     word (%0000000 << 9) | do_pop  
{11000010}     word (%0000011 << 9) | do_jnz  
{11000011}     word (%0000011 << 9) | do_jmp  
{11000100}     word (%0000011 << 9) | do_cnz  
{11000101}     word (%0000000 << 9) | do_push  
{11000110}     word (%0000001 << 9) | do_adi  
{11000111}     word (%0000000 << 9) | do_rst  
{11001000}     word (%0000000 << 9) | do_rz  
{11001001}     word (%0000000 << 9) | do_ret  
{11001010}     word (%0000011 << 9) | do_jz  
{11001011}     word (%0000000 << 9) | run  
{11001100}     word (%0000011 << 9) | do_cz  
{11001101}     word (%0000011 << 9) | do_call  
{11001110}     word (%0000001 << 9) | do_aci  
{11001111}     word (%0000000 << 9) | do_rst  
{11010000}     word (%0000000 << 9) | do_rnc  
{11010001}     word (%0000000 << 9) | do_pop  
{11010010}     word (%0000011 << 9) | do_jnc  
{11010011}     word (%0000001 << 9) | do_out  
{11010100}     word (%0000011 << 9) | do_cnc  
{11010101}     word (%0000000 << 9) | do_push  
{11010110}     word (%0000001 << 9) | do_sui  
{11010111}     word (%0000000 << 9) | do_rst  
{11011000}     word (%0000000 << 9) | do_rc  
{11011001}     word (%0000000 << 9) | run  
{11011010}     word (%0000011 << 9) | do_jc  
{11011011}     word (%0000001 << 9) | do_in  
{11011100}     word (%0000011 << 9) | do_cc  
{11011101}     word (%0000000 << 9) | run  
{11011110}     word (%0000001 << 9) | do_sbi  
{11011111}     word (%0000000 << 9) | do_rst  
{11100000}     word (%0000000 << 9) | do_rpo  
{11100001}     word (%0000000 << 9) | do_pop  
{11100010}     word (%0000011 << 9) | do_jpo  
{11100011}     word (%0000000 << 9) | do_xthl  
{11100100}     word (%0000011 << 9) | do_cpo  
{11100101}     word (%0000000 << 9) | do_push  
{11100110}     word (%0000001 << 9) | do_ani  
{11100111}     word (%0000000 << 9) | do_rst  
{11101000}     word (%0000000 << 9) | do_rpe  
{11101001}     word (%0000000 << 9) | do_pchl  
{11101010}     word (%0000011 << 9) | do_jpe  
{11101011}     word (%0000000 << 9) | do_xchg  
{11101100}     word (%0000011 << 9) | do_cpe  
{11101101}     word (%0000000 << 9) | run  
{11101110}     word (%0000001 << 9) | do_xri  
{11101111}     word (%0000000 << 9) | do_rst  
{11110000}     word (%0000000 << 9) | do_rp  
{11110001}     word (%0000000 << 9) | do_pop  
{11110010}     word (%0000011 << 9) | do_jp  
{11110011}     word (%0000000 << 9) | do_di  
{11110100}     word (%0000011 << 9) | do_cp  
{11110101}     word (%0000000 << 9) | do_push  
{11110110}     word (%0000001 << 9) | do_ori  
{11110111}     word (%0000000 << 9) | do_rst  
{11111000}     word (%0000000 << 9) | do_rm  
{11111001}     word (%0000000 << 9) | do_sphl  
{11111010}     word (%0000011 << 9) | do_jm  
{11111011}     word (%0000000 << 9) | do_ei  
{11111100}     word (%0000011 << 9) | do_cm  
{11111101}     word (%0000000 << 9) | run  
{11111110}     word (%0000001 << 9) | do_cpi  
{11111111}     word (%0000000 << 9) | do_rst  


          